#!/bin/sh
sudo rm -fr /Library/Receipts/Neverball_Uninstall_Script.pkg /Library/Receipts/Neverball.pkg /Library/Receipts/Neverputt.pkg /Library/Receipts/Neverball_Data.pkg /Library/Receipts/Neverball_Docs.pkg /Library/Receipts/Neverball_Uninstaller.pkg /Library/Application\ Support/Neverball\ Data /Applications/Neverball.app /Applications/Neverputt.app
